@@ -178,6 +178,44 @@ static switches_t dswitches[]= {
     { "CODEPAGE", OFFS(codepage), T_UINT }
 };
 
+#if (defined  HAVE_TOUPPER_L || defined HAVE_STRNCASECMP_L)
+static locale_t C=NULL;
+
+static void init_canon(void) {
+    if(C == NULL)
+	C = newlocale(LC_CTYPE_MASK, "C", NULL);
+}
+#endif
+
+#ifdef HAVE_TOUPPER_L
+static int canon_drv(int drive) {
+    int ret;
+    init_canon();
+    ret = toupper_l(drive, C);
+    return ret;
+}
+#else
+static int canon_drv(int drive) {
+    return toupper(drive);
+}
+#endif
+
+#ifdef HAVE_STRNCASECMP_L
+static int cmp_tok(const char *a, const char *b, int len) {
+    init_canon();
+    return strncasecmp_l(a, b, len, C);
+}
+#else
+static int cmp_tok(const char *a, const char *b, int len) {
+    return strncasecmp(a, b, len);
+}
+#endif
+
+
+static char ch_canon_drv(char drive) {
+    return (char) canon_drv( (unsigned char) drive);
+}
+
 static void maintain_default_drive(char drive)
 {
     if(default_drive == ':')
@@ -206,7 +244,10 @@ static void syntax(const char *msg, int thisLine)
     fprintf(stderr,"Syntax error at line %d ", lastTokenLinenumber);
     if(drive) fprintf(stderr, "for drive %c: ", drive);
     if(token) fprintf(stderr, "column %ld ", (long)(token - buffer));
-    fprintf(stderr, "in file %s: %s\n", filename, msg);
+    fprintf(stderr, "in file %s: %s", filename, msg);
+    if(errno != 0)
+	fprintf(stderr, " (%s)", strerror(errno));
+    fprintf(stderr, "\n");
     exit(1);
 }
 
@@ -218,23 +259,37 @@ static void get_env_conf(void)
     for(i=0; i< sizeof(global_switches) / sizeof(*global_switches); i++) {
 	s = getenv(global_switches[i].name);
 	if(s) {
-	    if(global_switches[i].type == T_INT)
+	    errno = 0;
+	    switch(global_switches[i].type) {
+	    case T_INT:
 		* ((int *)global_switches[i].address) = strtoi(s,0,0);
-	    if(global_switches[i].type == T_UINT)
+		break;
+	    case T_UINT:
 		* ((unsigned int *)global_switches[i].address) = strtoui(s,0,0);
-	    if(global_switches[i].type == T_UINT8)
+		break;
+	    case T_UINT8:
 		* ((uint8_t *)global_switches[i].address) = strtou8(s,0,0);
-	    if(global_switches[i].type == T_UINT16)
+		break;
+	    case T_UINT16:
 		* ((uint16_t *)global_switches[i].address) = strtou16(s,0,0);
-	    else if (global_switches[i].type == T_STRING)
+		break;
+	    case T_STRING:
 		* ((char **)global_switches[i].address) = s;
+		break;
+	    }
+	    if(errno != 0) {
+		fprintf(stderr, "Bad number %s for %s (%s)\n", s,
+			global_switches[i].name,
+			strerror(errno));
+		exit(1);
+	    }
 	}
     }
 }
 
 static int mtools_getline(void)
 {
-    if(!fp || !fgets(buffer, MAX_LINE_LEN, fp))
+    if(!fp || !fgets(buffer, MAX_LINE_LEN+1, fp))
 	return -1;
     linenumber++;
     pos = buffer;
@@ -280,7 +335,7 @@ static char *get_next_token(void)
 static int match_token(const char *template)
 {
     return (strlen(template) == token_length &&
-	    !strncasecmp(template, token, token_length));
+	    !cmp_tok(template, token, token_length));
 }
 
 static void expect_char(char c)
@@ -319,6 +374,8 @@ static unsigned long get_unumber(unsigned long max)
     skip_junk(1);
     last = pos;
     n=strtoul(pos, &pos, 0);
+    if(errno)
+	syntax("bad number", 0);
     if(last == pos)
 	syntax("numeral expected", 0);
     if(n > max)
@@ -336,6 +393,8 @@ static int get_number(void)
     skip_junk(1);
     last = pos;
     n=(int) strtol(pos, &pos, 0);
+    if(errno)
+	syntax("bad number", 0);
     if(last == pos)
 	syntax("numeral expected", 0);
     pos++;
@@ -348,7 +407,7 @@ static void purge(char drive, int fn)
 {
     unsigned int i, j;
 
-    drive = ch_toupper(drive);
+    drive = ch_canon_drv(drive);
     for(j=0, i=0; i < cur_devs; i++) {
 	if(devices[i].drive != drive ||
 	   devices[i].file_nr == fn)
@@ -425,7 +484,7 @@ static void finish_drive_clause(void)
     if(!trusted && (devices[cur_dev].misc_flags & PRIV_FLAG)) {
 	fprintf(stderr,
 		"Warning: privileged flag ignored for drive %c: defined in file %s\n",
-		toupper(devices[cur_dev].drive), filename);
+		canon_drv(devices[cur_dev].drive), filename);
 	devices[cur_dev].misc_flags &= ~PRIV_FLAG;
     }
     trusted = 0;
@@ -476,7 +535,6 @@ static int set_openflags(struct device *dev)
 static int set_misc_flags(struct device *dev)
 {
     unsigned int i;
-
     for(i=0; i < sizeof(misc_flags) / sizeof(*misc_flags); i++) {
 	if(match_token(misc_flags[i].name)) {
 	    flag_mask |= misc_flags[i].flag;
@@ -521,7 +579,7 @@ static int set_def_format(struct device *dev)
     return 1;
 }
 
-static int parse_one(int privilege);
+static void parse_all(int privilege);
 
 void set_cmd_line_image(char *img) {
   char *ofsp;
@@ -555,10 +613,22 @@ void set_cmd_line_image(char *img) {
     lastTokenLinenumber = 0;
     pos = buffer;
     token = 0;
-    while (parse_one(0));
+    parse_all(0);
   }
 }
 
+void check_number_parse_errno(char c, const char *oarg, char *endptr) {
+    if(endptr && *endptr) {
+	fprintf(stderr, "Bad number %s\n", oarg);
+	exit(1);
+    }
+    if(errno) {
+	fprintf(stderr, "Bad number %s for -%c (%s)\n", oarg,
+		c, strerror(errno));
+	exit(1);
+    }
+}
+
 static uint16_t tou16(int in, const char *comment) {
     if(in > UINT16_MAX) {
 	fprintf(stderr, "Number of %s %d too big\n", comment, in);
@@ -621,7 +691,7 @@ static void parse_old_device_line(char drive)
 	devices[cur_dev].heads = 0;
     }
 	
-    devices[cur_dev].drive = ch_toupper(devices[cur_dev].drive);
+    devices[cur_dev].drive = ch_canon_drv(devices[cur_dev].drive);
     maintain_default_drive(devices[cur_dev].drive);
     if (!(devices[cur_dev].name = strdup(name))) {
 	printOom();
@@ -663,7 +733,7 @@ static int parse_one(int privilege)
 	memset((char*)(devices+cur_dev), 0, sizeof(*devices));
 	trusted = privilege;
 	flag_mask = 0;
-	devices[cur_dev].drive = ch_toupper(token[0]);
+	devices[cur_dev].drive = ch_canon_drv(token[0]);
 	maintain_default_drive(devices[cur_dev].drive);
 	expect_char(':');
 	return 1;
@@ -686,6 +756,12 @@ static int parse_one(int privilege)
     return 1;
 }
 
+static void parse_all(int privilege) {
+    errno=0;
+    while (parse_one(privilege));
+}
+
+
 static int parse(const char *name, int privilege)
 {
     if(fp) {
@@ -709,7 +785,7 @@ static int parse(const char *name, int privilege)
     token = 0;
     cur_dev = -1; /* no current device */
 
-    while(parse_one(privilege));
+    parse_all(privilege);
     finish_drive_clause();
     fclose(fp);
     filename = NULL;
@@ -772,7 +848,7 @@ void mtoolstest(int argc, char **argv, int type  UNUSEDP)
     char drive='\0';
 
     if(argc > 1 && argv[1][0] && argv[1][1] == ':') {
-	drive = ch_toupper(argv[1][0]);
+	drive = ch_canon_drv(argv[1][0]);
     }
 
     for (dev=devices; dev->name; dev++) {
